"""Utilities for debugging this library."""

import typing as t

import httpx


class ProcessTransportBase(httpx.HTTPTransport):
    """A transport that provides hooks for processing requests and responses."""

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        """Handle a request."""
        try:
            response = super().handle_request(request)
        except Exception as e:
            return self.process_exception(request, e)
        else:
            return self.process_response(request, response)

    def process_exception(
        self, request: httpx.Request, exception: Exception
    ) -> httpx.Response:
        """
        Process a request that raised an exception.

        This can re-raise the exception, log it, or otherwise handle it.
        It can also transmute the exception into a response, if desired.
        """
        raise exception

    def process_response(
        self, request: httpx.Request, response: httpx.Response
    ) -> httpx.Response:
        """
        Process a response.

        This can log the response, modify it, or otherwise handle it.
        It can also transmute the response into an exception, if desired.
        """
        return response


class DumpDebugTransport(ProcessTransportBase):
    """A transport that prints the request and response."""

    _out: t.TextIO

    def __init__(self, out: t.TextIO, *args, **kwargs):
        """Initialize the transport."""
        super().__init__(*args, **kwargs)
        self._out = out

    def process_response(
        self, request: httpx.Request, response: httpx.Response
    ) -> httpx.Response:
        """Print the request and response."""
        self._out.write("# Request:\n")
        self._out.write(f"{request.method} {request.url}\n")
        self._out.write("# Headers:\n")
        for k, v in request.headers.items():
            self._out.write(f"{k}: {v}\n")
        self._out.write("# Body:\n")
        self._out.write(f"{request.content.decode()}\n")
        self._out.write("# Response:\n")
        self._out.write(f"Status: {response.status_code}\n")
        self._out.write("# Headers:\n")
        for k, v in response.headers.items():
            self._out.write(f"{k}: {v}\n")
        self._out.write("# Body:\n")
        _ = response.read()
        self._out.write(f"{response.text}\n")
        return response


class CurlDebugTransport(ProcessTransportBase):
    """A transport that prints the request as a CURL command."""

    _out: t.TextIO

    def __init__(self, out: t.TextIO, *args, **kwargs):
        """Initialize the transport."""
        super().__init__(*args, **kwargs)
        self._out = out

    def process_response(
        self, request: httpx.Request, response: httpx.Response
    ) -> httpx.Response:
        """Print the request as a CURL command."""
        SHOW_HEADERS = {"content-type"}
        self._out.write(f"curl -X {request.method} {request.url} \\\n")
        for k, v in request.headers.items():
            if k in SHOW_HEADERS:
                self._out.write(f"--header '{k}: {v}' \\\n")
        # This is one massive hack to make the output more readable.

        dbg_content = (
            request.content.decode().replace("\\n", "\n").replace("><", ">\n<")
        )
        self._out.write(f"--data '{dbg_content}'\n")
        return response



import re

_UNIT_TYPES = {
    "apartment": "APT",
    "basement": "BSM",
    "box #": "BOX",
    "building": "BLD",
    "department": "DEP",
    "floor": "FL",
    "front": "FRN",
    "hanger": "HNG",
    "lobby": "LBB",
    "lot": "LOT",
    "lower": "LOW",
    "office": "OFC",
    "penthouse": "PH",
    "pier": "PIE",
    "poll": "POL",
    "rear": "REA",
    "room": "RM",
    "side": "SID",
    "slip": "SLI",
    "space": "SPC",
    "stop": "STO",
    "suite": "STE",
    "trailer": "TRLR",
    "unit": "UNI",
    "upper": "UPP",
    "cabin": "CBN",
    "hub": "HUB",
    "student mailing center": "SMC",
    "townhouse": "TH",
}


_RE_BARE_NUMBER = re.compile(r"^#?(\d+)$")
_RE_BARE_UNIT_NUMBER = re.compile(r"^(\w+)\.? (\d+)$")
_RE_TRAILING_NUMBER = re.compile(r"^(.*) #(\d+)$")
_RE_TRAILING_UNIT_NUMBER = re.compile(r"^(.*) (\w+)\.? (\d+)$")


def get_unit_type_number(
    address1: str, address2: str | None
) -> tuple[str | None, str | None]:
    """
    Normalize a provided potentially multi-line address.

    Specifically, attempt to identify a unit type/number from the address,
    if a known type is found.

    If found, return a tuple of the unit type and number. Otherwise, return
    a tuple of None.
    """
    return None, None



"""Tools for working with voter registration data."""

from .ga import GeorgiaCheckRegistrationTool
from .mi import MichiganCheckRegistrationTool
from .pa.check import PennsylvaniaCheckRegistrationTool
from .pa.client import PennsylvaniaAPIClient
from .tool import CheckRegistrationTool
from .wi import WisconsinCheckRegistrationTool
from .zipcodes import get_state

_CHECK_TOOLS: dict[str, type[CheckRegistrationTool]] = {
    "GA": GeorgiaCheckRegistrationTool,
    "MI": MichiganCheckRegistrationTool,
    "PA": PennsylvaniaCheckRegistrationTool,
    "WI": WisconsinCheckRegistrationTool,
}


def get_check_tool(
    *, zipcode: str | None = None, state: str | None = None
) -> CheckRegistrationTool | None:
    """Return a voter registration tool for the given ZIP code or state."""
    if state is None:
        if zipcode is None:
            raise ValueError("Must provide either a ZIP code or state")
        state = get_state(zipcode)

    if state is None:
        return None

    tool_class = _CHECK_TOOLS.get(state)
    return tool_class() if tool_class else None


# TODO FUTURE: consider whether there's *any* kind of common interface
# between OVR API clients for different states. For now, we just have
# Pennsylvania, so there's no point in trying to generalize.


__all__ = [
    "get_check_tool",
    "PennsylvaniaAPIClient",
]



import datetime
import io
import typing as t
from base64 import b64decode, b64encode
from enum import Enum
from urllib.parse import urlencode

import httpx
import pydantic as p
import pydantic_xml as px
from PIL import Image

from .counties import CountyChoice, get_county_choice
from .errors import (
    APIError,
    InvalidRegistrationError,
    UnexpectedResponseError,
    get_error_class,
)

STAGING_URL = "https://paovrwebapi.beta.vote.pa.gov/SureOVRWebAPI/api/ovr"
PRODUCTION_URL = "https://paovrwebapi.vote.pa.gov/SureOVRWebAPI/api/ovr"

try:
    from lxml.etree import ElementBase as XmlElement  # type: ignore
    from lxml.etree import fromstring as xml_fromstring  # type: ignore
    from lxml.etree import tostring as xml_tostring  # type: ignore
except ImportError:
    from xml.etree.ElementTree import Element as XmlElement
    from xml.etree.ElementTree import fromstring as xml_fromstring
    from xml.etree.ElementTree import tostring as xml_tostring


# -----------------------------------------------------------------------------
# Pydantic validators & serializers for common data types found in the API
# -----------------------------------------------------------------------------

# Response dates from the PA API come in this format
PA_API_RESPONSE_DATE_FORMAT = "%m/%d/%Y"


def parse_response_date(v: str | datetime.date) -> datetime.date:
    """Parse dates in the PA API format."""
    if isinstance(v, datetime.date):
        return v
    return datetime.datetime.strptime(v, PA_API_RESPONSE_DATE_FORMAT).date()


def serialize_response_date(v: datetime.date) -> str:
    """Serialize dates in the PA API format."""
    return v.strftime(PA_API_RESPONSE_DATE_FORMAT)


PAResponseDate: t.TypeAlias = t.Annotated[
    datetime.date,
    p.PlainSerializer(serialize_response_date),
    p.BeforeValidator(parse_response_date),
]


# Response date times from the PA API come in this format
# 'Jun 26 2024  7:43PM' <-- may have extra spaces
PA_API_RESPONSE_DT_FORMAT = "%b %d %Y  %I:%M%p"


def parse_response_dt(v: str | datetime.datetime) -> datetime.datetime:
    """Parse dates in the PA API format."""
    if isinstance(v, datetime.datetime):
        return v
    return datetime.datetime.strptime(v, PA_API_RESPONSE_DT_FORMAT)


def serialize_response_dt(v: datetime.datetime) -> str:
    """Serialize dates in the PA API format."""
    return v.strftime(PA_API_RESPONSE_DT_FORMAT)


PAResponseDateTime: t.TypeAlias = t.Annotated[
    datetime.datetime,
    p.PlainSerializer(serialize_response_dt),
    p.BeforeValidator(parse_response_dt),
]


# Request dates sent to the PA API should be in this format
PA_API_REQUEST_DATE_FORMAT = "%Y-%m-%d"


def parse_request_date(v: str | datetime.date) -> datetime.date:
    """Parse dates in the PA API format."""
    if isinstance(v, datetime.date):
        return v
    return datetime.datetime.strptime(v, PA_API_REQUEST_DATE_FORMAT).date()


def serialize_request_date(v: datetime.date) -> str:
    """Serialize dates in the PA API format."""
    return v.strftime(PA_API_REQUEST_DATE_FORMAT)


PARequestDate: t.TypeAlias = t.Annotated[
    datetime.date,
    p.PlainSerializer(serialize_request_date),
    p.BeforeValidator(parse_request_date),
]


# At least this one is consistent!
PA_API_TIME_FORMAT = "%I:%M %p"


def parse_time(v: str | datetime.time) -> datetime.time:
    """Parse times in the PA API format."""
    if isinstance(v, datetime.time):
        return v
    return datetime.datetime.strptime(v, PA_API_TIME_FORMAT).time()


def serialize_time(v: datetime.time) -> str:
    """Serialize times in the PA API format."""
    return v.strftime(PA_API_TIME_FORMAT)


PATime: t.TypeAlias = t.Annotated[
    datetime.time, p.PlainSerializer(serialize_time), p.BeforeValidator(parse_time)
]


def title_case(v: str) -> str:
    """Ensure the string is title-cased."""
    return v.title()


TitleCaseStr: t.TypeAlias = t.Annotated[str, p.AfterValidator(title_case)]


def parse_phone_number(v: str) -> str:
    """Transform phone numbers to the PA API format, ###-###-####."""
    # Remove all non-numeric characters from the phone number.
    digits = "".join(filter(str.isdigit, v))
    # If there are 11 digits, check if the first digit is a '1' and remove it.
    if len(digits) == 11:
        if digits[0] != "1":
            raise ValueError("Phone number must start with '1' if it has 11 digits.")
        digits = digits[1:]
    if len(digits) != 10:
        raise ValueError("Phone number must have 10 digits.")
    # Format the phone number as ###-###-####.
    return f"{digits[:3]}-{digits[3:6]}-{digits[6:]}"


PhoneNumber: t.TypeAlias = t.Annotated[str, p.BeforeValidator(parse_phone_number)]


def validate_bit(v: int | str | bool) -> bool:
    """Validate that the value is a bit (0 or 1)."""
    if isinstance(v, bool):
        return v
    if isinstance(v, str):
        if v not in {"0", "1"}:
            raise ValueError("Bit value must be '0' or '1'.")
        return v == "1"
    if v not in {0, 1}:
        raise ValueError("Bit value must be 0 or 1.")
    return bool(v)


def bit_serializer(v: bool) -> int:
    """
    Serialize a boolean value as an integer (0 or 1).

    This is used for lots of boolean values in the PA API, which are
    represented as integers (0 or 1) in the XML.
    """
    return int(v)


Bit: t.TypeAlias = t.Annotated[
    bool, p.PlainSerializer(bit_serializer), p.BeforeValidator(validate_bit)
]


def validate_true_bit(v: int | str | bool) -> t.Literal[True]:
    """Validate that the value is a true bit (1)."""
    as_bit = validate_bit(v)
    if not as_bit:
        raise ValueError("Bit value must resolve to `True`.")
    return True


TrueBit: t.TypeAlias = t.Annotated[
    t.Literal[True],
    p.PlainSerializer(bit_serializer),
    p.BeforeValidator(validate_true_bit),
]


def _binaryio_to_image(bio: t.BinaryIO) -> Image.Image:
    """Convert a binary file-like object to a PIL Image."""
    # Open the image from the file-like object.
    try:
        image = Image.open(bio)
        image.load()
    except Exception as e:
        raise ValueError("Invalid image data.") from e
    return image


def _bytes_to_image(b: bytes) -> Image.Image:
    """Convert a byte string to a PIL Image."""
    return _binaryio_to_image(io.BytesIO(b))


def data_url_to_image(data_url: str) -> Image.Image:
    """Convert a data: URL to a PIL Image."""
    # Split the data URL into its components.
    parts = data_url.split(",")
    if len(parts) != 2:
        raise ValueError("Invalid data: URL format.")

    if not parts[0].startswith("data:image/"):
        raise ValueError("Invalid data: URL format.")

    if not parts[0].endswith(";base64"):
        raise ValueError("Invalid data: URL format.")

    # Extract the data and decode it.
    try:
        data = parts[1].encode("utf-8")
        decoded = b64decode(data)
    except Exception as e:
        raise ValueError("Invalid data: URL format.") from e

    return _bytes_to_image(decoded)


def image_to_data_url(image: Image.Image) -> str:
    """Convert a PIL Image to a data: URL."""
    # Create a file-like object to store the image data.
    bio = io.BytesIO()
    image.save(bio, format="PNG", dpi=(75, 75))
    # Encode the image data as a base64 string.
    encoded = b64encode(bio.getvalue()).decode("utf-8")
    # Return the data: URL.
    return f"data:image/png;base64,{encoded}"


SIGNATURE_IMAGE_SIZE = (180, 60)


def validate_signature_image(
    v: str | bytes | t.BinaryIO | Image.Image,
) -> str:
    """
    Validate that the value is a valid signature image.

    The validator accepts a string, bytes, a binary file-like object, or a PIL
    Image object. If a string is provided, it is assumed to be a valid
    data: URL for an image. If bytes are provided, they are assumed to be the
    raw image data. If a file-like object is provided, it is read and its
    contents are used as the image data. If a PIL Image object is provided, it
    is treated as the image itself.

    Raw image data is converted to a canonical format, specifically a black
    and white PNG that is 180x60 pixels in size, marked as 75 DPI. The PA
    documentation says it wants < 98% white pixels and < 90% black pixels,
    but we mostly ignore that for now.

    The validator returns a string, which is the data: URL for the image.
    """
    image: Image.Image

    # Perform the basic conversion
    if isinstance(v, str):
        image = data_url_to_image(v)
    elif isinstance(v, bytes):
        image = _bytes_to_image(v)
    elif isinstance(v, t.BinaryIO) or isinstance(v, io.BytesIO):
        image = _binaryio_to_image(v)
    elif isinstance(v, Image.Image):
        image = v
    else:
        raise ValueError("Invalid signature image data kind.")

    # Resize the image to 180x60 pixels, make it black and white.
    resized = (
        image.convert("L")
        .point(lambda p: 255 if p > 56 else 0, mode="1")
        .resize(SIGNATURE_IMAGE_SIZE)
    )

    # Convert the image to a data: URL.
    return image_to_data_url(resized)


SignatureImage: t.TypeAlias = t.Annotated[
    str, p.BeforeValidator(validate_signature_image)
]


# -----------------------------------------------------------------------------
# Generic response type
# -----------------------------------------------------------------------------


class APIResponse(px.BaseXmlModel, tag="RESPONSE", frozen=True):
    """
    A generic response from the Pennsylvania OVR API.

    This is a simple response that contains a status code and a message. Not
    all API calls return this response structure, but some do.

    It is also the response structure used on *any* error response, including
    for API calls that would normally return a different response type.
    """

    application_id: str | None = px.element(alias="APPLICATIONID", default=None)
    """The application ID, if one was created."""

    application_date: PAResponseDateTime | None = px.element(
        alias="APPLICATIONDATE", default=None
    )
    """The application date, if one was created."""

    signature: str | None = px.element(alias="SIGNATURE", default=None)
    """The signature, if one was created."""

    error_codes: tuple[str, ...] | None = px.element(alias="ERROR", default=None)
    """
    The error codes, if an error occurred.

    If batch mode is set to ALL_ERRORS, this will contain all errors found, 
    otherwise it will contain at most one.
    """

    @p.field_validator("error_codes", mode="before")
    def validate_error_codes(cls, value: t.Any) -> tuple[str, ...] | None:
        """Ensure that error codes are always a tuple."""
        if value is None:
            return None
        if isinstance(value, str):
            return tuple(value.split(","))
        # We assume there's something iterable here.
        try:
            # First, collapse that thing back into a string.
            value = ",".join(value)
            # Now split it back into a tuple. (Yeah, I know; sorry.)
            return tuple(value.split(","))
        except Exception as e:
            # Nope, not even iterable.
            raise ValueError("Error codes must be a string or an iterable.") from e

    @property
    def error_code(self) -> str | None:
        """Return the first error code, if any."""
        if self.error_codes is None:
            return None
        return self.error_codes[0]

    def has_error(self) -> bool:
        """Check if the response indicates an error."""
        # On occasion, the API will return a response with neither an error code
        # nor an application ID. We treat this as a registration error.
        return (self.error_code is not None) or (self.application_id is None)

    def get_error_class(self) -> t.Type[APIError]:
        """Get the error class for the error code, or the default error class."""
        if self.error_code is None:
            return InvalidRegistrationError
        return get_error_class(self.error_code)

    def raise_for_error(self) -> None:
        """Raise an exception if the response indicates an error."""
        if self.has_error():
            error_class = self.get_error_class()
            raise error_class(f"codes: {self.error_codes}")


# -----------------------------------------------------------------------------
# *SETUP Response Types
# -----------------------------------------------------------------------------


class SetupOption(px.BaseXmlModel, frozen=True):
    """Base class for all options returned by the *SETUP API calls."""

    pass


class Suffix(SetupOption, frozen=True):
    """A suffix for a name, like 'jr'."""

    code: str = px.element(alias="NameSuffixCode")
    """The code for the suffix (like 'jr')"""

    description: str = px.element(alias="NameSuffixDescription")
    """The human-readable description of the suffix (like 'JR.')."""


class Race(SetupOption, frozen=True):
    """The type of a person's race, like 'white'."""

    code: str = px.element(alias="RaceCode")
    """The code for the race (like 'W')."""

    description: TitleCaseStr = px.element(alias="RaceDescription")
    """The human-readable description for the race (like 'WHITE')."""


class UnitType(SetupOption, tag="UnitTypes", frozen=True):
    """The type of an address unit, like 'building'."""

    code: str = px.element(alias="UnitTypesCode")
    """The code for the unit type (like 'BLD')."""

    description: TitleCaseStr = px.element(alias="UnitTypesDescription")
    """The human-readable description for the unit type (like 'BUILDING')."""


class AssistanceType(SetupOption, frozen=True):
    """The type of assistance needed, like 'hearing impaired'."""

    code: str = px.element(alias="AssistanceTypeCode")
    """The code for the assistance type (like 'HI')."""

    description: str = px.element(alias="AssistanceTypeDescription")
    """
    The human-readable description for the assistance type.

    (like 'I am deaf or hard of hearing').
    """


class Gender(SetupOption, frozen=True):
    """The type of a person's gender, like 'female'."""

    code: str = px.element(alias="GenderCode")
    """The code for the gender (like 'U')."""

    description: str = px.element(alias="GenderDescription")
    """The human-readable description for the gender (like 'Unknown')."""


class PoliticalParty(SetupOption, frozen=True):
    """The type of a person's political party, like 'democratic'."""

    code: str = px.element(alias="PoliticalPartyCode")
    """The code for the political party (like 'D')."""

    description: str = px.element(alias="PoliticalPartyDescription")
    """The human-readable description for the political party (like 'Democratic')."""


class MailinAddressType(SetupOption, tag="MailinAddressTypes", frozen=True):
    """The type of a mail-in address, like 'residential'."""

    code: str = px.element(alias="MailinAddressTypesCode")
    """The code for the mail-in address type (like 'R')."""

    description: str = px.element(alias="MailinAddressTypesDescription")
    """
    The human-readable description for the mail-in address type.
     
    (like 'Residential Address').
    """


class County(SetupOption, frozen=True):
    """A single county in Pennsylvania, like 'Adams'."""

    code: int = px.element(alias="countyID")
    """The ID for the county (like '2290')."""

    name: TitleCaseStr | None = px.element(alias="Countyname", default=None)
    """
    The name of the county (like 'ADAMS').
    
    SETUP calls may return a single county with no name.
    """


class State(SetupOption, tag="States", frozen=True):
    """A state in the United States, like 'Pennsylvania'."""

    code: str = px.element(alias="Code")
    """The code for the state (like 'PA')."""

    name: str = px.element(alias="CodesDescription")
    """The human-readable description for the state (like 'Pennsylvania')."""


class SetupResponse(
    px.BaseXmlModel, tag="NewDataSet", search_mode="unordered", frozen=True
):
    """
    A full response from the *SETUP API call.

    The setup call returns a variety of mappings and other data that can be used
    to build a voter registration or mail-in ballot application.
    """

    suffixes: tuple[Suffix, ...] = px.element(tag="Suffix")
    """All available suffixes for a name."""

    races: tuple[Race, ...] = px.element(tag="Race")
    """All available race options."""

    unit_types: tuple[UnitType, ...] = px.element(tag="UnitTypes")
    """All available unit types for an address (like 'Apartment')."""

    assistance_types: tuple[AssistanceType, ...] = px.element(tag="AssistanceType")
    """All available assistance types (like 'I am deaf or hard of hearing')."""

    genders: tuple[Gender, ...] = px.element(tag="Gender")
    """All available gender options."""

    political_parties: tuple[PoliticalParty, ...] = px.element(tag="PoliticalParty")
    """All available political parties."""

    mailin_address_types: tuple[MailinAddressType, ...] = px.element(
        tag="MailinAddressTypes"
    )
    """All available mail-in address types (like 'Residence')."""

    counties: tuple[County, ...] = px.element(tag="County")
    """All available counties in Pennsylvania."""

    states: tuple[State, ...] = px.element(tag="States")
    """All available states in the United States."""

    next_registration_deadline: PAResponseDate = px.wrapped(
        "NextVRDeadline/NextVRDeadline"
    )
    """The next voter registration deadline."""

    next_election: PAResponseDate = px.wrapped("NextElection/NextElection")
    """The next election date."""

    ovr_app_declaration: str = px.wrapped("Text_OVRApplnDeclaration/Text")
    """HTML text containing a legal declaration for the online application."""

    ovr_assistance_declaration: str = px.wrapped(
        "Text_OVRApplnAssistanceDeclaration/Text"
    )
    """HTML text containing a legal declaration for assisted OVR application."""

    mail_in_declaration: str = px.wrapped("Text_OVRMailInApplnDeclaration/Text")
    """HTML text containing a legal declaration for the mail-in ballot application."""

    election_name: TitleCaseStr = px.wrapped("Text_OVRMailInElectionName/ElectionName")
    """The name of the next election."""

    mail_in_completion_time: PATime = px.wrapped("Text_OVRMailInApplnComplTime/Time")
    """The expected completion time-of-day for a mail-in ballot application."""

    mail_in_received_time: PATime = px.wrapped(
        "Text_OVRMailInBallotRecvdTime/RecvdTime"
    )
    """The cutoff time for receiving a mail-in ballot on a given day."""


# -----------------------------------------------------------------------------
# Municipalities Response
# -----------------------------------------------------------------------------


class Municipality(px.BaseXmlModel, tag="Municipality", frozen=True):
    """A single municipality in Pennsylvania, like 'Barnett Township'."""

    mun_type: int = px.element(alias="MunicipalityType")
    """The numerical type of the municipality."""

    id: str | None = px.element(alias="MunicipalityID", default=None)
    """The ID for the municipality (like 'MN01')."""

    name: TitleCaseStr | None = px.element(alias="MunicipalityIDname", default=None)
    """The name of the municipality (like 'BARNETT TOWNSHIP')."""

    county_id: int = px.element(alias="CountyID")
    """The ID for the county (like '2290')."""

    county_name: TitleCaseStr | None = px.element(alias="CountyName", default=None)
    """The name of the county (like 'ADAMS')."""


class MunicipalitiesResponse(px.BaseXmlModel, tag="OVRLookupData", frozen=True):
    """A response from the GETMUNICIPALITIES API call."""

    municipalities: tuple[Municipality, ...]
    """A list of municipalities in the given county."""


# -----------------------------------------------------------------------------
# Error Values Response
# -----------------------------------------------------------------------------


class ErrorValue(px.BaseXmlModel, tag="MessageText", frozen=True):
    """A single error value from the GETERRORVALUES API call."""

    code: str = px.element(alias="ErrorCode")
    """The code for the error value (like 'VR_WAPI_InvalidAccessKey')."""

    text: str = px.element(alias="ErrorText")
    """The human-readable description of the error value."""


class ErrorValuesResponse(px.BaseXmlModel, tag="OVRLookupData", frozen=True):
    """A response from the GETERRORVALUES API call."""

    errors: tuple[ErrorValue, ...]
    """A mapping from error code to error message."""


# -----------------------------------------------------------------------------
# Languages Response
# -----------------------------------------------------------------------------


class Language(px.BaseXmlModel, tag="Languages", frozen=True):
    """A single language in the Pennsylvania OVR API."""

    code: str = px.element(alias="LanguageCode")
    """The code for the language (like 'LANGENG')."""

    name: str = px.element(alias="Language")
    """The human-readable name for the language (like 'English')."""


class LanguagesResponse(px.BaseXmlModel, tag="OVRLookupData", frozen=True):
    """A response from the GETLANGUAGES API call."""

    languages: tuple[Language, ...]
    """A list of languages available in the API."""


# -----------------------------------------------------------------------------
# Set Application Request
# -----------------------------------------------------------------------------


NSMAP = {"": "OVRexternaldata"}


class BatchMode(int, Enum):
    """Enumeration of batch numbers for voter registration applications."""

    FIRST_ERROR_ONLY = 0
    """Request that the API returns exactly one error, if any is found."""

    ALL_ERRORS = 1
    """Request that the API returns all errors found in the application."""


class SuffixChoice(str, Enum):
    """
    Enumeration of suffixes for a person's name.

    CONSIDER: the API returns a list of valid <Suffix> in the *SETUP response.
    These seem quite stable, so we simply hard-code them here.
    """

    NONE = ""
    FIRST = "I"
    SECOND = "II"
    THIRD = "III"
    FOURTH = "IV"
    FIFTH = "V"
    SIXTH = "VI"
    SEVENTH = "VII"
    JUNIOR = "JR"
    SENIOR = "SR"


class PoliticalPartyChoice(str, Enum):
    """
    Enumeration of political parties in Pennsylvania.

    CONSIDER: the API returns a list of valid <PoliticalParty> in the *SETUP
    response. These seem quite stable, so we simply hard-code them here.
    """

    DEMOCRATIC = "D"
    REPUBLICAN = "R"
    GREEN = "GR"
    LIBERTARIAN = "LN"
    NO_AFFILIATION = "NF"
    OTHER = "OTH"


class GenderChoice(str, Enum):
    """
    Enumeration of possible genders.

    CONSIDER: the API returns a list of valid <Gender> in the *SETUP response.
    These seem quite stable, so we simply hard-code them here.
    """

    FEMALE = "F"
    MALE = "M"
    UNKNOWN = "U"


class RaceChoice(str, Enum):
    """
    Enumeration of possible races/ethnicities.

    CONSIDER: the API returns a list of valid <Race> in the *SETUP response.
    These seem quite stable, so we simply hard-code them here.
    """

    ASIAN = "A"
    BLACK_OR_AFRICAN_AMERICAN = "B"
    HISPANIC_OR_LATINO = "H"
    NATIVE_AMERICAN_OR_ALASKAN_NATIVE = "I"
    NATIVE_HAWAIIAN_OR_OTHER_PACIFIC_ISLANDER = "P"
    OTHER = "O"
    TWO_OR_MORE_RACES = "T"
    WHITE = "W"


class RegistrationKind(int, Enum):
    """Enumeration of registration kinds for a voter registration application."""

    CHANGE = 0
    """The application is a change of registration."""

    NEW = 1
    """The application is a new registration."""


class UnitTypeChoice(str, Enum):
    """
    Enumeration of possible address unit types.

    CONSIDER: the API returns a list of valid <UnitTypes> in the *SETUP response.
    These seem quite stable, so we simply hard-code them here.
    """

    APARTMENT = "APT"
    BASEMENT = "BSM"
    BOX = "BOX"
    BUILDING = "BLD"
    DEPARTMENT = "DEP"
    FLOOR = "FL"
    FRONT = "FRN"
    HANGER = "HNG"
    LOBBY = "LBB"
    LOT = "LOT"
    LOWER = "LOW"
    OFFICE = "OFC"
    PENTHOUSE = "PH"
    PIER = "PIE"
    POLL = "POL"
    REAR = "REA"
    ROOM = "RM"
    SIDE = "SID"
    SLIP = "SLI"
    SPACE = "SPC"
    STOP = "STO"
    SUITE = "STE"
    TRAILER = "TRL"  # XXX the API also returns a separate TRLR tuple. Ugh.
    UNIT = "UNI"
    UPPER = "UPP"
    CABIN = "CBN"
    HUB = "HUB"
    STUDENT_MAILING_CENTER = "SMC"
    TOWNHOUSE = "TH"


class AssistanceTypeChoice(str, Enum):
    """
    Enumeration of possible assistance types.

    CONSIDER: the API returns a list of valid <AssistanceType> in the *SETUP
    response. These seem quite stable, so we simply hard-code them here.
    """

    HARD_OF_HEARING = "HI"  # "I am deaf or hard of hearing"
    VISION_IMPAIRED = "VI"  # "I am blind or have difficulty seeing"
    USES_WHEELCHAIR = "WC"  # "I use a wheelchair"
    PHYSICAL_DISABILITY = "PD"  # "I have a physical disability"
    READING = "IL"  # "I need help reading"
    LANGUAGE = "LN"  # "I do not speak English well"


class MailInAddressTypeChoice(str, Enum):
    """
    Enumeration of possible mail-in address types.

    CONSIDER: the API returns a list of valid <MailinAddressTypes> in the *SETUP
    response. These seem quite stable, so we simply hard-code them here.
    """

    RESIDENTIAL = "R"
    MAILING = "M"
    ALTERNATE = "A"


class VoterBatch(px.BaseXmlModel, frozen=True):
    """01 - A batch mode for a voter registration application."""

    # XXX setting BatchMode.FIRST_ERROR_ONLY results in the staging
    # endpoint returning empty responses. But *the exact same* application
    # data works fine with BatchMode.ALL_ERRORS. This must be a bug in the API.
    batch: BatchMode = px.element(default=BatchMode.ALL_ERRORS)


class VoterName(px.BaseXmlModel, frozen=True):
    """02 - Personal information for a voter registration application."""

    first_name: str = px.element(tag="FirstName", max_length=30)
    middle_name: str | None = px.element(tag="MiddleName", default=None, max_length=30)
    last_name: str = px.element(tag="LastName", max_length=30)
    suffix: SuffixChoice = px.element(tag="TitleSuffix", default=SuffixChoice.NONE)


class VoterEligibility(px.BaseXmlModel, frozen=True):
    """03 - Eligibility information for a voter registration application."""

    is_us_citizen: TrueBit = px.element(tag="united-states-citizen")
    will_be_18: TrueBit = px.element(tag="eighteen-on-election-day")

    @p.model_validator(mode="after")
    def validate_age(self) -> t.Self:
        """Ensure that the voter will be 18 on election day."""
        if not self.will_be_18:
            raise ValueError("Voter must be 18 on election day.")
        return self

    @p.model_validator(mode="after")
    def validate_citizenship(self) -> t.Self:
        """Ensure that the voter is a US citizen."""
        if not self.is_us_citizen:
            raise ValueError("Voter must be a US citizen.")
        return self


class VoterReason(px.BaseXmlModel, frozen=True):
    """

    Manage the reasons for a voter registration application.

    This combines two sections of the API documentation spreadsheet:

    04 - Reason for a voter registration application, and
    11 - Voting information has changed

    since proper validation requires the combination of both.
    """

    registration_kind: RegistrationKind = px.element(tag="isnewregistration")
    """Whether the application is a new registration, or a change."""

    is_name_change: Bit = px.element(tag="name-update", default=False)
    """Whether the application includes a name change."""

    is_address_change: Bit = px.element(tag="address-update", default=False)
    """Whether the application includes an address change."""

    is_party_change: Bit = px.element(tag="ispartychange", default=False)
    """Whether the application includes a party change."""

    is_federal_voter: Bit = px.element(tag="isfederalvoter", default=False)
    """
    Whether the application is for a Federal or State employee registering
    in their county of last residence.

    This is a bit confusing, but it's what the PA docs say. Moreover, this
    can apparently be *combined* with either new *or* change registrations.
    Okay!
    """

    @p.model_validator(mode="after")
    def validate_reasons(self) -> t.Self:
        """
        If the application is a new registration, no reasons should be selected.

        If the application is a change of registration, at least one reason
        should be selected.
        """
        any_reasons = any(
            (
                self.is_name_change,
                self.is_address_change,
                self.is_party_change,
            )
        )
        if self.registration_kind == RegistrationKind.NEW:
            if any_reasons:
                raise ValueError(
                    "No reasons should be selected for a new registration."
                )
        else:
            if not any_reasons:
                raise ValueError(
                    "At least one reason should be selected for a change regs."
                )
        return self

    voter_reg_number: str | None = px.element(
        "voterregnumber", default=None, max_length=12
    )
    """
    The voter registration number, if the application is a change.

    This is not strictly required for a change application, but it may
    help the API identify the voter.
    """

    previous_last_name: str | None = px.element(
        "previousreglastname", default=None, max_length=30
    )
    """If the application includes a name change, the previous last name."""

    previous_first_name: str | None = px.element(
        "previousregfirstname", default=None, max_length=30
    )
    """If the application includes a name change, the previous first name."""

    previous_middle_name: str | None = px.element(
        "previousregmiddlename", default=None, max_length=30
    )
    """If the application includes a name change, the previous middle name."""

    previous_address: str | None = px.element(
        "previousregaddress", default=None, max_length=100
    )
    """If the application includes an address change, the previous address."""

    previous_city: str | None = px.element(
        "previousregcity", default=None, max_length=30
    )
    """If the application includes an address change, the previous city."""

    previous_state: str | None = px.element(
        "previousregstate", default=None, max_length=2
    )
    """If the application includes an address change, the previous state."""

    previous_zip5: str | None = px.element("previousregzip", default=None, max_length=5)
    """If the application includes an address change, the previous ZIP code."""

    previous_county: CountyChoice | None = px.element("previousregcounty", default=None)
    """If the application includes an address change, the previous county."""

    previous_year: str | None = px.element(
        "previousregyear", default=None, min_length=4, max_length=4
    )
    """If the application includes an address change, the previous year."""

    @p.model_validator(mode="after")
    def validate_previous_name(self) -> t.Self:
        """Ensure that previous names are set if the app includes a name change."""
        values = (
            self.previous_last_name,
            self.previous_first_name,
        )
        if self.is_name_change:
            if not all(values):
                raise ValueError("Previous name fields must be set for a name change.")
        else:
            if any(values):
                raise ValueError("No previous name fields should be set.")
        return self

    @p.model_validator(mode="after")
    def validate_previous_address(self) -> t.Self:
        """Ensure that previous addr fields are set if the app includes a change."""
        values = (
            self.previous_address,
            self.previous_city,
            self.previous_zip5,
            self.previous_county,
        )
        if self.is_address_change:
            if not all(values):
                raise ValueError(
                    "Previous address fields must be set for an address change."
                )
        else:
            if any(values):
                raise ValueError("No previous address fields should be set.")
        return self


class VoterAbout(px.BaseXmlModel, frozen=True):
    """05 - Information about the voter for a voter registration application."""

    birth_date: PARequestDate = px.element(tag="DateOfBirth")

    gender: GenderChoice | None = px.element(tag="Gender", default=None)
    """Optional gender. (Some PA docs use the term 'sex', some use 'gender')."""

    race: RaceChoice | None = px.element("Ethnicity", default=None)
    """Optional ethnicity. (Some PA docs use the term 'race', some 'ethnicity')."""

    phone: PhoneNumber | None = px.element(
        tag="Phone", default=None, min_length=12, max_length=12
    )
    email: p.EmailStr | None = px.element(tag="Email", max_length=50, default=None)


class VoterAddress(px.BaseXmlModel, frozen=True):
    """06 - Address information for a voter registration application."""

    address: str = px.element(tag="streetaddress", max_length=40)
    address_2: str | None = px.element(
        tag="streetaddress2", default=None, max_length=40
    )
    unit_type: UnitTypeChoice | None = px.element(tag="unittype", default=None)
    unit_number: str | None = px.element(tag="unitnumber", default=None, max_length=15)
    city: str = px.element(tag="city", max_length=35)
    zip5: str = px.element(tag="zipcode", max_length=5, min_length=5)

    # FUTURE: make `unit_type` and `unit_number` computed fields, based on the
    # contents of the address lines themselves. Otherwise it's just weirdness?

    @p.field_validator("zip5", mode="after")
    def validate_zip5_has_county(cls, value: str) -> str:
        """Ensure that the ZIP code has a corresponding county."""
        if get_county_choice(value) is None:
            raise ValueError(f"Unknown county for ZIP code {value}")
        return value

    @px.computed_element(tag="county")  # type: ignore
    def county(self) -> CountyChoice:
        """Return the county for the given ZIP code."""
        choice = get_county_choice(self.zip5)
        if choice is None:
            raise ValueError(f"Unknown county for ZIP code {self.zip5}")
        return choice


class VoterMailingAddress(px.BaseXmlModel, frozen=True):
    """07 - Mailing address information for a voter registration application."""

    no_street_permanent: Bit = px.element(
        tag="donthavePermtOrResAddress", default=False
    )
    """If True, voter does not have a street address or permanent residence."""

    mailing_address: str | None = px.element(
        tag="mailingaddress", default=None, max_length=40
    )
    """Optional mailing address for the voter if no_street_permanent is set."""

    mailing_city: str | None = px.element(
        tag="mailingcity", default=None, max_length=35
    )
    """Optional mailing city for the voter if no_street_permanent is set."""

    mailing_state: str | None = px.element(
        tag="mailingstate", default=None, max_length=2
    )
    """Optional mailing state for the voter if no_street_permanent is set."""

    mailing_zipcode: str | None = px.element(
        tag="mailingzipcode", default=None, max_length=10
    )
    """
    Optional mailing ZIP code for the voter if no_street_permanent is set.
    
    Unlike the primary ZIP code, this field can be longer than 5 characters.
    :shrug-emoji: 
    """

    @p.model_validator(mode="after")
    def validate_mailing_address(self) -> t.Self:
        """Ensure that all mailing address fields are set if any are set."""
        values = (
            self.mailing_address,
            self.mailing_city,
            self.mailing_state,
            self.mailing_zipcode,
        )
        if self.no_street_permanent:
            if not all(values):
                raise ValueError("All mailing fields must be set if any are set.")
        else:
            if any(values):
                raise ValueError(
                    "No mailing fields should be set if no_street_permanent is False."
                )
        return self


class VoterIdentification(px.BaseXmlModel, frozen=True):
    """08 - Identification information for a voter registration application."""

    drivers_license: str | None = px.element(
        "drivers-license", default=None, max_length=8
    )
    """The voter's PA driver's license number *or* PennDOT id card number."""

    ssn4: str | None = px.element("ssn4", default=None, max_length=4)
    """The last 4 digits of the voter's Social Security Number."""

    @px.computed_element(tag="donthavebothDLandSSN")  # type: ignore
    def no_identification(self) -> Bit:
        """Return True if the voter has neither a driver's license nor an SSN."""
        return self.drivers_license is None and self.ssn4 is None

    signature: SignatureImage | None = px.element("signatureimage", default=None)
    """The voter's signature image."""

    @px.computed_element(tag="continueAppSubmit")  # type: ignore
    def continue_submit(self) -> Bit:
        """
        Return True if the application can continue a previous failed app.

        XXX The API documentation is a little fuzzy here; we should clarify. It
        specifically says that this is required to complete the application
        if a previous PennDOT DL or ID fails.
        """
        return self.drivers_license is None

    @p.model_validator(mode="after")
    def validate_identification(self) -> t.Self:
        """Ensure that API rules are met."""
        if self.drivers_license is not None:
            if any((self.ssn4, self.signature)):
                raise ValueError(
                    "SSNs and signatures are not allowed if a DL is provided."
                )
        elif self.ssn4 is not None:
            if self.signature is not None:
                raise ValueError("Signatures are not allowed if an SSN is provided.")
        elif self.signature is None:
            raise ValueError(
                "A signature is required if no identification is provided."
            )
        elif (
            self.drivers_license is None
            and self.ssn4 is None
            and self.signature is None
        ):
            raise ValueError("At least one form of identification must be supplied.")
        return self


class VoterPoliticalParty(px.BaseXmlModel, frozen=True):
    """09 - Political party information for a voter registration application."""

    political_party: PoliticalPartyChoice = px.element(tag="politicalparty")
    """The political party of the voter."""

    other_party: str | None = px.element(tag="otherpoliticalparty", default=None)
    """The name of the political party if 'Other' is selected."""

    # Validate that `other_party` is set if `political_party` is `OTHER`.
    @p.model_validator(mode="after")
    def validate_other_party(self) -> t.Self:
        """Ensure that `other_party` is set if `political_party` is `OTHER`."""
        if (self.political_party == PoliticalPartyChoice.OTHER) ^ (
            self.other_party is not None
        ):
            raise ValueError(
                "`other_party` must (only) be set if `political_party` is `OTHER`"
            )
        return self


class VoterAssistance(px.BaseXmlModel, frozen=True):
    """10 - Assistance information for a voter registration application."""

    require_help_to_vote: Bit = px.element(tag="needhelptovote", default=False)
    """Whether the voter requires assistance to vote."""

    # If require_help_to_vote is True, this must also be set.
    assistance_type: AssistanceTypeChoice | None = px.element(
        tag="typeofassistance", default=None
    )
    """The type of assistance needed by the voter."""

    # Despite the fact that not all assistance types require a language, the API
    # requires this to be set if require_help_to_vote is True.
    preferred_language: str | None = px.element(tag="preferredlanguage", default=None)
    """The preferred language of the voter."""

    @p.model_validator(mode="after")
    def validate_assistance(self) -> t.Self:
        """Ensure that all assistance fields are set if any are set."""
        values = (self.assistance_type, self.preferred_language)
        if self.require_help_to_vote:
            if not all(values):
                raise ValueError("All assistance fields must be set if any are set.")
        else:
            if any(values):
                raise ValueError("No assistance fields should be set if not needed.")
        return self


class VoterChangedInfo(px.BaseXmlModel, frozen=True):
    """11 - Changed information for a voter registration application."""

    # The implementation for this section of the API docs is
    # handled by VoterReason, since the validation required combines
    # both section 04 and section 11.

    pass


class VoterDeclaration(px.BaseXmlModel, frozen=True):
    """12 - Declaration information for a voter registration application."""

    confirm_declaration: TrueBit = px.element(tag="declaration1")
    """
    The voter's confirmation of the declaration.

    The declaration is provided by the *SETUP API call as `ovr_app_declaration`.
    """


class VoterHelpWithForm(px.BaseXmlModel, frozen=True):
    """13 - Help with form information for a voter registration application."""

    assistant_name: str | None = px.element(
        tag="assistedpersonname", max_length=100, default=None
    )
    """The name of the person who assisted the voter in filling out the form."""

    assistant_address: str | None = px.element(
        tag="assistedpersonAddress", max_length=100, default=None
    )
    """The address of the person who assisted the voter in filling out the form."""

    assistant_phone: PhoneNumber | None = px.element(
        tag="assistedpersonphone", max_length=15, default=None
    )

    assistant_declaration: TrueBit | None = px.element(
        tag="assistancedeclaration2", default=None
    )
    """
    The assistant's confirmation of the declaration.

    The declaration is provided by the *SETUP API call as `ovr_assistance_declaration`.
    """

    @p.model_validator(mode="after")
    def ensure_assistant_fields(self) -> t.Self:
        """Ensure that all assistant fields are set if any are set."""
        values = (
            self.assistant_name,
            self.assistant_address,
            self.assistant_phone,
            self.assistant_declaration,
        )
        if any(values) and not all(values):
            raise ValueError("All assistant fields must be set if any are set.")

        return self


class VoterPollWorker(px.BaseXmlModel, frozen=True):
    """14 - Poll worker information for a voter registration application."""

    be_poll_worker: Bit | None = px.element(tag="ispollworker", default=None)
    """If True, the voter wants to be a poll worker on election day."""

    be_interpreter: Bit | None = px.element(tag="bilingualinterpreter", default=None)
    """If True, the voter wants to be a bilingual interpreter on election day."""

    interpreter_language: str | None = px.element(
        tag="pollworkerspeaklang", default=None, max_length=50
    )
    """The language the voter speaks if they want to be a bilingual interpreter."""

    @p.model_validator(mode="after")
    def validate_interpreter(self) -> t.Self:
        """Ensure that interpreter fields are set if any are set."""
        if self.be_interpreter:
            if not self.interpreter_language:
                raise ValueError(
                    "An interpreter language is required if `be_interpreter`"
                )
        else:
            if self.interpreter_language:
                raise ValueError(
                    "`interpreter_language` is only allowed if `be_interpreter`"
                )
        return self


class VoterSecondEmailID(px.BaseXmlModel, frozen=True):
    """15 - Second email ID information for a voter registration application."""

    alternate_email: p.EmailStr | None = px.element(
        tag="secondEmail", max_length=50, default=None
    )


class VoterTransferPermanentStatusFlag(px.BaseXmlModel, frozen=True):
    """16 - Transfer Permanent Status Flag information for a voter reg application."""

    transfer_permanent_status: Bit | None = px.element(
        tag="istransferpermanent", default=None
    )
    """
    If True, the voter would like their 'permanent status' transferred to 
    their new county of residence when doing a change of address registration.
    """


class VoterMailInBallot(px.BaseXmlModel, frozen=True):
    """17 - Mail-In Ballot information for a voter registration application."""

    is_mail_in: Bit | None = px.element(tag="ismailinballot", default=None)
    """If True, the voter would like to receive a mail-in ballot."""

    mail_in_address_type: MailInAddressTypeChoice | None = px.element(
        tag="mailinaddresstype", default=None
    )
    """The type of mail-in address the voter would like to use."""

    mail_in_address: str | None = px.element(
        tag="mailinballotaddr", default=None, max_length=40
    )
    """The street address where a mail-in ballot should be mailed."""

    mail_in_city: str | None = px.element(tag="mailincity", default=None, max_length=35)
    """The city where a mail-in ballot should be mailed."""

    mail_in_state: str | None = px.element(
        tag="mailinstate", default=None, max_length=2
    )
    """The state where a mail-in ballot should be mailed."""

    mail_in_zipcode: str | None = px.element(
        tag="mailinzipcode", default=None, max_length=10
    )
    """The ZIP code where a mail-in ballot should be mailed."""

    mail_in_ward: str | None = px.element(tag="mailinward", default=None, max_length=50)
    """The ward where a mail-in ballot should be mailed."""

    mail_in_lived_since: PARequestDate | None = px.element(
        tag="mailinlivedsince", default=None
    )
    """
    The year the voter started living at the mail-in address.
    
    TODO: verify that this is actually what's expected here; the documentation
    is not at all helpful.
    """

    mail_in_declaration: TrueBit | None = px.element(
        tag="mailindeclaration", default=None
    )
    """
    The voter's confirmation of the mail-in ballot declaration.

    The declaration is provided by the *SETUP API call as `mail_in_declaration`.
    """

    @p.model_validator(mode="after")
    def validate_mail_in_address(self) -> t.Self:
        """Ensure that all mail-in address fields are set if any are set."""
        values = (
            self.mail_in_address_type,
            self.mail_in_address,
            self.mail_in_city,
            self.mail_in_state,
            self.mail_in_zipcode,
            # mail_in_ward is not required according to the API docs
            self.mail_in_lived_since,
            self.mail_in_declaration,
        )
        if self.is_mail_in:
            if not all(values):
                raise ValueError("All mail-in fields must be set if any are set.")
        else:
            if any(values):
                raise ValueError(
                    "No mail-in fields should be set if `is_mail_in` is False."
                )
        return self


class VoterApplicationRecord(
    VoterMailInBallot,
    VoterTransferPermanentStatusFlag,
    VoterSecondEmailID,
    VoterPollWorker,
    VoterHelpWithForm,
    VoterDeclaration,
    VoterChangedInfo,
    VoterAssistance,
    VoterPoliticalParty,
    VoterIdentification,
    VoterMailingAddress,
    VoterAddress,
    VoterAbout,
    VoterReason,
    VoterEligibility,
    VoterName,
    VoterBatch,
    tag="record",
    nsmap=NSMAP,
    frozen=True,
    search_mode="unordered",
):
    """A record for a voter registration application."""

    # NOTE: the inheritance order is important here. The API expects
    # the elements to be in a specific order, so we need to ensure that
    # the elements are defined in the correct order in the class definition.

    pass


class VoterApplication(
    px.BaseXmlModel,
    tag="APIOnlineApplicationData",
    nsmap=NSMAP,
    frozen=True,
):
    """A full voter registration application for the Pennsylvania OVR API."""

    record: VoterApplicationRecord
    """Silly, needless nesting in the API request structure. So be it."""


# -----------------------------------------------------------------------------
# API Verbs (what the PA API documentation calls "Actions")
# -----------------------------------------------------------------------------


class Action(str, Enum):
    """
    Enumeration of actions that can be taken against the Pennsylvania OVR API.

    The API has a single URL endpoint, which accepts a variety of so-called
    "actions" that determine what the API should do.
    """

    GET_APPLICATION_SETUP = "GETAPPLICATIONSETUP"
    """Get the possible values for a voter reg + optional mail-in ballot app."""

    GET_BALLOT_APPLICATION_SETUP = "GETBALLOTAPPLICATIONSETUP"
    """Get the possible values for a mail-in ballot app."""

    GET_LANGUAGES = "GETLANGUAGES"
    """Get the available languages for the PA OVR API."""

    GET_XML_TEMPLATE = "GETXMLTEMPLATE"
    """Get XML tags and format for voter registration + optional mail-in ballot app."""

    GET_BALLOT_XML_TEMPLATE = "GETBALLOTXMLTEMPLATE"
    """Get XML tags and format for mail-in ballot app."""

    GET_ERROR_VALUES = "GETERRORVALUES"
    """Get the possible error values for the PA OVR API."""

    GET_MUNICIPALITIES = "GETMUNICIPALITIES"
    """Get the available municipalities in a given county."""

    SET_APPLICATION = "SETAPPLICATION"
    """Submit a voter registration + optional mail-in ballot app."""

    SET_BALLOT_APPLICATION = "SETBALLOTAPPLICATION"
    """Submit a mail-in ballot app."""


# -----------------------------------------------------------------------------
# Client implementation
# -----------------------------------------------------------------------------


class PennsylvaniaAPIClient:
    """Client for speaking with the Pennsylvania online voter (OVR) API."""

    api_url: str
    api_key: str
    language: int
    _client: httpx.Client

    def __init__(
        self,
        api_url: str,
        api_key: str,
        *,
        language: int = 0,
        timeout: float = 5.0,
        # Lower-level parameter for test and debug purposes
        _transport: httpx.BaseTransport | None = None,
    ):
        """Create a new client for the Pennsylvania OVR API."""
        self.api_url = api_url
        self.api_key = api_key
        self.language = language
        mounts = {"all://": _transport} if _transport else None
        self._client = httpx.Client(mounts=mounts, timeout=timeout)

    @classmethod
    def staging(
        cls, api_key: str, *, language: int = 0, timeout: float = 5.0, **kwargs
    ):
        """Create a new client for the Pennsylvania OVR API in staging."""
        return cls(STAGING_URL, api_key, language=language, timeout=timeout, **kwargs)

    @classmethod
    def production(
        cls, api_key: str, *, language: int = 0, timeout: float = 5.0, **kwargs
    ):
        """Create a new client for the Pennsylvania OVR API in production."""
        return cls(
            PRODUCTION_URL, api_key, language=language, timeout=timeout, **kwargs
        )

    def build_url(self, action: Action, params: dict | None = None) -> str:
        """Build a URL for the given action and parameters."""
        query = {
            "AuthKey": self.api_key,
            "action": action.value,
            "Language": self.language,
            **(params or {}),
        }

        query_prefix = {f"sysparm_{k}": v for k, v in query.items()}
        encoded = urlencode(query_prefix)
        return f"{self.api_url}?JSONv2&{encoded}"

    def _raw_get(self, action: Action, params: dict | None = None) -> str:
        """Perform a raw GET request to the Pennsylvania OVR API."""
        url = self.build_url(action, params)
        response = self._client.get(url, headers={"Cache-Control": "no-cache"})
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise UnexpectedResponseError() from e
        return response.json()

    def _raw_post(
        self,
        action: Action,
        data: XmlElement | str,  # type: ignore
        params: dict | None = None,
    ) -> str:
        """Perform a raw POST request to the Pennsylvania OVR API."""
        url = self.build_url(action)
        if isinstance(data, XmlElement):
            data_str = xml_tostring(data, encoding="unicode")  # type: ignore
            assert isinstance(data_str, str)
            # XXX this is no fun -- the API doesn't *really*
            # accept xml, because if it did, this would not be necessary.
            # Instead, the API accepts a *very specific variant* of XML
            # where the namespaces are just-so. And I can't figure out how
            # to produce that variant from pydantic-xml. So we're doing this.
            data_str = data_str.replace("ns0:", "").replace(":ns0=", "=")
        else:
            data_str = data

        assert isinstance(data_str, str)

        data_jsonable = {"ApplicationData": data_str}
        print("TODO DAVE remove this: DATA jsonable: ", data_jsonable)
        response = self._client.post(
            url,
            json=data_jsonable,
            headers={"Cache-Control": "no-cache"},
        )
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise UnexpectedResponseError() from e
        return response.json()

    def _get(self, action: Action, params: dict | None = None) -> XmlElement:
        """Perform a GET request to the Pennsylvania OVR API."""
        raw = self._raw_get(action, params)
        try:
            return xml_fromstring(raw)
        except Exception as e:
            raise UnexpectedResponseError() from e

    def _post(
        self, action: Action, data: XmlElement | str, params: dict | None = None
    ) -> XmlElement:
        """Perform a POST request to the Pennsylvania OVR API."""
        raw = self._raw_post(action, data, params)
        try:
            return xml_fromstring(raw)
        except Exception as e:
            raise UnexpectedResponseError() from e

    def _raise_if_error(self, xml_response: XmlElement) -> None:
        """Raise an APIError or subclass if the xml_response indicates an error."""
        # Regardless of the specific invocation we just made, the API *always*
        # returns a "generic" API response when there's an error code.
        try:
            # ET.indent(xml_response)
            # ET.dump(xml_response)
            api_response = APIResponse.from_xml_tree(xml_response)
            api_response.raise_for_error()
        except px.ParsingError:
            # The response was not a "generic" response so there was no error.
            pass

    def invoke(
        self,
        action: Action,
        data: XmlElement | str | None = None,
        params: dict | None = None,
    ) -> XmlElement:
        """
        Invoke an action on the Pennsylvania OVR API.

        Look for errors in the response and raise an exception if one is found.
        """
        response_data = (
            self._post(action, data, params)
            if data is not None
            else self._get(action, params)
        )
        self._raise_if_error(response_data)
        return response_data

    def get_application_setup(self) -> SetupResponse:
        """Get the possible values for a voter reg + optional mail-in ballot app."""
        data = self.invoke(Action.GET_APPLICATION_SETUP)
        return SetupResponse.from_xml_tree(data)

    def get_ballot_application_setup(self) -> SetupResponse:
        """Get the possible values for a mail-in ballot app."""
        data = self.invoke(Action.GET_BALLOT_APPLICATION_SETUP)
        return SetupResponse.from_xml_tree(data)

    def get_languages(self) -> LanguagesResponse:
        """Get the available languages for the PA OVR API."""
        data = self.invoke(Action.GET_LANGUAGES)
        return LanguagesResponse.from_xml_tree(data)

    def get_xml_template(self) -> XmlElement:
        """Get XML tags and format for voter reg + optional mail-in ballot app."""
        return self.invoke(Action.GET_XML_TEMPLATE)

    def get_ballot_xml_template(self) -> XmlElement:
        """Get XML tags and format for mail-in ballot app."""
        return self.invoke(Action.GET_BALLOT_XML_TEMPLATE)

    def get_error_values(self) -> ErrorValuesResponse:
        """Get the possible error values for the PA OVR API."""
        data = self.invoke(Action.GET_ERROR_VALUES)
        return ErrorValuesResponse.from_xml_tree(data)

    def get_municipalities(self, county: str) -> MunicipalitiesResponse:
        """Get the available municipalities in a given county."""
        data = self.invoke(Action.GET_MUNICIPALITIES, params={"County": county})
        return MunicipalitiesResponse.from_xml_tree(data)

    def set_application(self, application: VoterApplication) -> APIResponse:
        """Submit a voter registration + optional mail-in ballot app."""
        xml_tree = application.to_xml_tree()
        data = self.invoke(Action.SET_APPLICATION, data=xml_tree)
        return APIResponse.from_xml_tree(data)

    def set_ballot_application(self, application: VoterApplication) -> APIResponse:
        """Submit a mail-in ballot app."""
        data = self.invoke(
            Action.SET_BALLOT_APPLICATION, data=application.to_xml_tree()
        )
        return APIResponse.from_xml_tree(data)



import datetime
import typing as t
from abc import ABC, abstractmethod

import pydantic as p


class CheckRegistrationDetails(p.BaseModel, frozen=True):
    """Details about a voter's registration status."""

    state_id: str
    """The voter's state ID number."""

    registration_date: datetime.date
    """The date the voter registered."""

    status: str
    """The voter's registration status (active, inactive, suspended, etc.)."""


class CheckRegistrationResult(p.BaseModel, frozen=True):
    """The result of a voter registration check."""

    registered: bool
    """True if the voter is or was at one point registered to vote."""

    details: CheckRegistrationDetails | None = None
    """Details about the voter's registration, if available."""


class SupportedFeatures(p.BaseModel, frozen=True):
    """Features supported by a voter tool."""

    details: bool
    """True if the tool provides detailed registration information."""


class CheckRegistrationTool(ABC):
    """Base class for working with voter registration data."""

    state: t.ClassVar[str]
    """The two-letter state abbreviation for this tool."""

    features: t.ClassVar[SupportedFeatures]
    """Features supported by this tool."""

    @abstractmethod
    def check_registration(
        self,
        first_name: str,
        last_name: str,
        zipcode: str,
        birth_date: datetime.date,
        details: bool = False,
    ) -> CheckRegistrationResult:
        """
        Check whether a voter is registered to vote.

        If details are requested and available, include them in the result.
        (Not all states provide detailed registration information.)

        Raises a CheckRegistrationError if there is an unexpected failure.
        """
        ...


# TODO: as we support online voter registration in more states, consider if
# there's anything like a unified interface to build for them. For now,
# let's not bother.



import datetime
import typing as t

import httpx
from bs4 import BeautifulSoup

from .errors import CheckRegistrationError, MultipleRecordsFoundError
from .soup_utils import find_attr_value
from .tool import (
    CheckRegistrationDetails,
    CheckRegistrationResult,
    CheckRegistrationTool,
    SupportedFeatures,
)

# ------------------------------------------------------------------------
# CheckRegistrationTool implementation for MI
# ------------------------------------------------------------------------


class MichiganCheckRegistrationTool(CheckRegistrationTool):
    """A tool for checking voter registration in Michigan."""

    state: t.ClassVar[str] = "MI"
    features: t.ClassVar[SupportedFeatures] = SupportedFeatures(details=True)

    SEARCH_BY_NAME_URL: t.ClassVar[str] = (
        "https://mvic.sos.state.mi.us/Voter/SearchByName"
    )

    def check_registration(
        self,
        first_name: str,
        last_name: str,
        zipcode: str,
        birth_date: datetime.date,
        details: bool = False,
    ) -> CheckRegistrationResult:
        """Check if a voter is registered in Michigan."""
        try:
            request = httpx.post(
                self.SEARCH_BY_NAME_URL,
                data={
                    "FirstName": first_name,
                    "LastName": last_name,
                    "NameBirthMonth": birth_date.strftime("%m"),
                    "NameBirthYear": birth_date.strftime("%Y"),
                    "ZipCode": zipcode,
                    "Dln": "",
                    "DlnBirthMonth": "",
                    "DlnBirthYear": "",
                    "DpaID": 0,
                    "Months": None,
                    "VoterNotFound": False,
                    "TransitionVoter": False,
                },
            )
            request.raise_for_status()
        except httpx.HTTPError as e:
            raise CheckRegistrationError("Failed to check voter registration.") from e

        # Parse the response
        soup = BeautifulSoup(request.text, "html.parser")

        # See if there are multiple voter records
        multiple_records_val = find_attr_value(
            soup, "input", "value", id="hfmultiplevoterrecords"
        )
        if (multiple_records_val or "").lower() == "true":
            raise MultipleRecordsFoundError()

        # Check if the voter was found
        voter_found_val = find_attr_value(soup, "input", "value", id="hfnotfound")
        if (voter_found_val or "").lower() == "true":
            return CheckRegistrationResult(registered=False)

        # Don't want details? We're done here.
        if not details:
            return CheckRegistrationResult(registered=True)

        # Get the registration details
        dpa_id_val = find_attr_value(soup, "input", "value", id="Voter_0__DpaID")
        if not dpa_id_val:
            raise CheckRegistrationError("Failed to find dpa id.")
        registration_date_val = find_attr_value(
            soup, "input", "value", id="Voter_0__EffectiveRegistrationDate"
        )
        if not registration_date_val:
            raise CheckRegistrationError("Failed to find registration date.")
        try:
            registration_date = datetime.datetime.strptime(
                registration_date_val.split(" ")[0], "%m/%d/%Y"
            ).date()
        except ValueError as e:
            raise CheckRegistrationError("Failed to parse registration date.") from e

        return CheckRegistrationResult(
            registered=True,
            details=CheckRegistrationDetails(
                state_id=dpa_id_val,
                registration_date=registration_date,
                status="active",  # TODO: are there alternatives?
            ),
        )



from bs4 import BeautifulSoup, Tag


def find_attr_value(
    soup: BeautifulSoup, tag_name: str, attr_name: str, **kwargs
) -> str | None:
    """Find the single string value of an attribute in a tag, if any."""
    result_set = soup.find_all(name=tag_name, **kwargs)
    if len(result_set) != 1:
        return None
    tag = result_set[0]
    if not isinstance(tag, Tag):
        return None
    attr_value = tag.get(attr_name)
    if not isinstance(attr_value, str):
        return None
    return attr_value



import datetime
import json
import sys
import typing as t
from abc import abstractmethod

import httpx
import pydantic as p

from .errors import CheckRegistrationError
from .tool import (
    CheckRegistrationDetails,
    CheckRegistrationResult,
    CheckRegistrationTool,
    SupportedFeatures,
)
from .zipcodes import get_county

# ------------------------------------------------------------------------
# Utilities for issuing requests to Georgia's check voter registration site
# ------------------------------------------------------------------------


class ActionResult(p.BaseModel, frozen=True):
    """Base class for results of actions on the GA voter reg site."""

    pass


class Action(p.BaseModel, frozen=True):
    """
    A single action taken on the GA voter reg site.

    A request to the site is composed of one or more actions.

    Each action has an identifier that is used to match it with its
    response data.
    """

    method: t.ClassVar[str]
    class_name: t.ClassVar[str] = "vr_MvpDashboardController"
    result_class: t.ClassVar[type[ActionResult]]

    cacheable: t.ClassVar[bool | None] = False
    is_continuation: t.ClassVar[bool | None] = False

    @abstractmethod
    def params(self) -> dict:
        """Return params used as part of the action data."""
        ...

    def to_data(self) -> dict:
        """Return data used as part of the larger request structure."""
        return {
            "id": str(self.__class__.__name__),
            "descriptor": "aura://ApexActionController/ACTION$execute",
            "callingDescriptor": "UNKNOWN",
            "params": {
                "namespace": "",
                "classname": self.class_name,
                "method": self.method,
                "params": self.params(),
                **({"cacheable": self.cacheable} if self.cacheable is not None else {}),
                **(
                    {"isContinuation": self.is_continuation}
                    if self.is_continuation is not None
                    else {}
                ),
            },
        }


class CheckContactExistMessage(p.BaseModel, frozen=True):
    """Message data returned from checking if a contact exists."""

    has_data: bool = p.Field(alias="hasData")
    is_multiple: bool = p.Field(alias="isMultiple")
    contact_id: str = p.Field(alias="id")
    instance_of: str = p.Field(alias="instanceOf")


class CheckContactExistResult(ActionResult, frozen=True):
    """The result of checking if a contact exists in the GA voter reg system."""

    success: bool
    error: bool
    message: CheckContactExistMessage


class CheckContactExistAction(Action, frozen=True):
    """An action to check if a contact exists in the GA voter reg system."""

    first_name: str
    last_name: str
    zipcode: str
    birth_date: datetime.date

    result_class: t.ClassVar[type[ActionResult]] = CheckContactExistResult
    class_name: t.ClassVar[str] = "vr_MvpLandingPageController"
    method: t.ClassVar[str] = "checkContactExist"
    cacheable: t.ClassVar[bool | None] = None
    is_continuation: t.ClassVar[bool | None] = None

    def params(self) -> dict:
        """Return params used as part of the action data."""
        county_name = get_county(self.zipcode)
        if county_name is None:
            raise ValueError(f"Unknown county for zipcode: {self.zipcode}")
        county_name = county_name.upper()
        return {
            "requestMap": {
                "firstInitial": self.first_name.upper()[0],
                "lastName": self.last_name.upper(),
                "county": county_name,
                "georgiaId": "",
                "recaptchaResponse": '""',
                "version": "",
                "birthDate": self.birth_date.strftime("%Y-%m-%d"),
                "ssn": None,
            },
        }


class ContactIDAction(Action, frozen=True):
    """A base action for actions that require a contact ID."""

    contact_id: str

    def params(self) -> dict:
        """Return params used as part of the action data."""
        return {"contactId": self.contact_id}


class GetCurrentContactInformationResult(ActionResult, frozen=True):
    """The result of getting the current contact information for a user."""

    alternate_id: str = p.Field(alias="contactId")
    full_name: str = p.Field(alias="name")  # ALL CAPS


class GetCurrentContactInformationAction(ContactIDAction, frozen=True):
    """An action to get the current contact information for a user."""

    method: t.ClassVar[str] = "getCurrentContactInformation"
    result_class: t.ClassVar[type[ActionResult]] = GetCurrentContactInformationResult


class GetPersonalInformationResult(ActionResult, frozen=True):
    """The result of getting the personal information for a user."""

    # There is a *ton* of information returned by this action. Pydantic
    # happily ignores undeclared fields, so we can just declare the ones
    # we care about for now. But there's a lot!

    registration_date: datetime.date = p.Field(alias="createdDate")

    @p.field_validator("registration_date", mode="before")
    def validate_dt_or_none(cls, value: str) -> datetime.date:
        """Validate the registration date."""
        return datetime.datetime.strptime(value, "%Y-%m-%d").date()

    email: str | None = None
    first_name: str = p.Field(alias="firstName")
    last_name: str = p.Field(alias="lastName")
    middle_name: str | None = p.Field(alias="middleName", default=None)
    gender: str
    has_maiden_name: bool = p.Field(alias="hasMaidenName")
    has_middle_name: bool = p.Field(alias="hasMiddleName")
    phone: str | None = None

    status: str  # Like "Active"

    voter_reg_number: str = p.Field(alias="voterRegistrationNumber")

    @property
    def is_active(self) -> bool:
        """Return whether the registration is active."""
        return self.status.lower() == "active"


class GetPersonalInformationAction(ContactIDAction, frozen=True):
    """An action to get the personal information for a user."""

    result_class: t.ClassVar[type[ActionResult]] = GetPersonalInformationResult
    method: t.ClassVar[str] = "getPersonalInformation"

    def params(self) -> dict:
        """Return params used as part of the action data."""
        return {"conId": self.contact_id, "isBallotInfoRequired": True}


ACTIONS: list[type[Action]] = [
    CheckContactExistAction,
    GetCurrentContactInformationAction,
    GetPersonalInformationAction,
]


class GARequest(p.BaseModel, frozen=True):
    """An arbitrary request to the GA voter reg site."""

    actions: t.Sequence[Action]

    @classmethod
    def _get_current_fwuid(cls) -> str:
        """Return the current 'framework UID' used on the GA reg site."""
        # TODO FUTURE: Improve this.
        #
        # This constant seems to indicate the current "version" of the site.
        #
        # If it goes out of date, requests formed by this script will fail.
        #
        # It looks like I can explicitly find  the current version by visiting
        # https://mvp.sos.ga.gov/s/ and looking for a script tag whose `src`
        # attribute contains an `fwuid` query parameter.
        #
        # For today, let's just hardcode the value.
        return "ZDROWDdLOGtXcTZqSWZiU19ZaDJFdzk4bkk0bVJhZGJCWE9mUC1IZXZRbmcyNDguMTAuNS01LjAuMTA"  # noqa: E501

    @classmethod
    def _get_current_appid(cls) -> str:
        """Return the current 'app markup ID' used on the GA reg site."""
        # Same story as with fwuid.
        return "7766VpxH8B5ZgC8Vrgi-bQ"

    @classmethod
    def _get_secondary_loader_id(cls) -> str:
        """Return the current 'secondary loader ID' used on the GA reg site."""
        # Same story as with fwuid.
        return "nSN3-Xh18FbrdCVGqsWZnw"

    @classmethod
    def _build_aura_context(cls) -> dict:
        """Return the current 'aura context' used on the GA reg site."""
        # Same story as with fwuid.
        return {
            "mode": "PROD",
            "fwuid": cls._get_current_fwuid(),
            "app": "siteforce:communityApp",
            "loaded": {
                "APPLICATION@markup://siteforce:communityApp": cls._get_current_appid(),
                "COMPONENT@markup://instrumentation:o11ySecondaryLoader": cls._get_secondary_loader_id(),  # noqa: E501
            },
            "dn": [],
            "globals": {},
            "uad": False,
        }

    def to_data(self) -> dict:
        """Return the data used as part of the larger request structure."""
        message = {"actions": [action.to_data() for action in self.actions]}
        return {
            "message": json.dumps(message),
            "aura.context": json.dumps(self._build_aura_context()),
            "aura.token": None,
            "aura.pageURI": "/s/",
        }


# ------------------------------------------------------------------------
# Utilities for processing responses from the GA voter reg site
# ------------------------------------------------------------------------


def _build_action_result(action_data: dict) -> ActionResult | None:
    """
    Process data for a single action result and build a typed result struct.

    On failure, return None.
    """
    action_id = action_data.get("id")
    if not action_id:
        return None
    action_class = next((a for a in ACTIONS if a.__name__ == action_id), None)
    if not action_class:
        return None
    state = action_data.get("state")
    if state != "SUCCESS":
        return None
    outer_return_value = action_data.get("returnValue")
    if not isinstance(outer_return_value, dict):
        return None
    inner_return_value = outer_return_value.get("returnValue")
    if not isinstance(inner_return_value, dict):
        return None
    try:
        result = action_class.result_class.model_validate(inner_return_value)
    except p.ValidationError as e:
        # CONSIDER: look into pydantic support for empty dict in union w/ dict
        if action_id != "CheckContactExistAction":
            print("ERROR: cannot parse: ", e, file=sys.stderr)
            print(json.dumps(inner_return_value, indent=2), file=sys.stderr)
        return None
    return result


def _build_action_results(action_datas: t.Sequence[dict]) -> t.Sequence[ActionResult]:
    """Extract action result data and return a tuple of results."""
    return tuple(
        action_result
        for action_data in action_datas
        if (action_result := _build_action_result(action_data))
    )


class GAResponse(p.BaseModel, frozen=True):
    """A response from the GA voter reg site."""

    results: t.Sequence[ActionResult]

    @classmethod
    def for_response_data(cls, data: t.Any) -> t.Self:
        """Create a response from arbitrary response data."""
        if not isinstance(data, dict):
            return cls(results=())
        action_datas = data.get("actions")
        if not isinstance(action_datas, list):
            return cls(results=())
        results = _build_action_results(action_datas)
        return cls(results=results)

    def result_for_action(self, action: Action) -> ActionResult | None:
        """Return the result for a given action."""
        return self.result_for_class(action.__class__)

    def result_for_class(self, action_class: type[Action]) -> ActionResult | None:
        """Return the result for a given action ID."""
        for result in self.results:
            if isinstance(result, action_class.result_class):
                return result
        return None


# ------------------------------------------------------------------------
# Utilities for issuing requests to the GA site
# ------------------------------------------------------------------------

GA_URL = "https://mvp.sos.ga.gov/s/sfsites/aura"


def _invoke_ga_endpoint(request: GARequest) -> httpx.Response:
    """Invoke an endpoint on the GA voter reg site."""
    final_url = f"{GA_URL}?aura.ApexAction.execute={len(request.actions)}"
    response = httpx.post(
        final_url,
        data=request.to_data(),
    )
    response.raise_for_status()
    return response


def make_ga_request(request: GARequest) -> GAResponse:
    """Make a request to the GA voter reg site."""
    response = _invoke_ga_endpoint(request)
    try:
        data = response.json()
    except Exception:
        print("INVALID JSON: ", response.text, file=sys.stderr)
        return GAResponse(results=())
    ga_response = GAResponse.for_response_data(data)
    return ga_response


def _check_contact_exist(
    first_name: str, last_name: str, zipcode: str, birth_date: datetime.date
) -> CheckContactExistResult | None:
    """
    Check if the user is registered to vote in Georgia.

    If so, return the voter's "contact ID". Otherwise, return None.
    """
    check_action = CheckContactExistAction(
        first_name=first_name,
        last_name=last_name,
        zipcode=zipcode,
        birth_date=birth_date,
    )
    request = GARequest(actions=(check_action,))
    response = make_ga_request(request)
    check_result = t.cast(
        CheckContactExistResult | None, response.result_for_action(check_action)
    )
    return check_result


def _get_contact_details(contact_id: str) -> GetPersonalInformationResult | None:
    """Get the details of a registered voter, by contact ID, in Georgia."""
    personal_action = GetPersonalInformationAction(contact_id=contact_id)
    request = GARequest(actions=(personal_action,))
    response = make_ga_request(request)
    personal_result = t.cast(
        GetPersonalInformationResult | None, response.result_for_action(personal_action)
    )
    return personal_result


# ------------------------------------------------------------------------
# CheckRegistrationTool implementation for GA
# ------------------------------------------------------------------------


class GeorgiaCheckRegistrationTool(CheckRegistrationTool):
    """A tool for checking voter registration in Georgia."""

    state: t.ClassVar[str] = "GA"
    features: t.ClassVar[SupportedFeatures] = SupportedFeatures(details=True)

    def check_registration(
        self,
        first_name: str,
        last_name: str,
        zipcode: str,
        birth_date: datetime.date,
        details: bool = False,
    ) -> CheckRegistrationResult:
        """Check whether a voter is registered in Georgia."""
        try:
            check_result = _check_contact_exist(
                first_name, last_name, zipcode, birth_date
            )
        except Exception as e:
            raise CheckRegistrationError("Error checking voter registration") from e

        if check_result is None:
            return CheckRegistrationResult(registered=False, details=None)

        if not check_result.success:
            raise CheckRegistrationError("Error checking voter registration")

        if not details:
            return CheckRegistrationResult(registered=True, details=None)

        contact_id = check_result.message.contact_id

        try:
            personal_result = _get_contact_details(contact_id)
        except Exception as e:
            raise CheckRegistrationError("Error checking voter registration") from e

        if personal_result is None:
            return CheckRegistrationResult(registered=True, details=None)

        return CheckRegistrationResult(
            registered=True,
            details=CheckRegistrationDetails(
                state_id=personal_result.voter_reg_number,
                registration_date=personal_result.registration_date,
                status=personal_result.status.lower(),
            ),
        )



import datetime
import typing as t

import httpx
import pydantic as p

from .errors import CheckRegistrationError, MultipleRecordsFoundError
from .tool import (
    CheckRegistrationDetails,
    CheckRegistrationResult,
    CheckRegistrationTool,
    SupportedFeatures,
)

# ------------------------------------------------------------------------
# Wisconsin API Data Structures
# ------------------------------------------------------------------------


class SearchResponseVoter(p.BaseModel, frozen=True):
    """An individual voter data structure for the search response."""

    voter_id: str = p.Field(alias="voterRegNumber")
    voter_status: str = p.Field(alias="voterStatusName")
    registration_date: datetime.date = p.Field(alias="registrationDate")

    @p.field_validator("voter_status", mode="after")
    def validate_status(cls, value: str) -> str:
        """Validate the voter status."""
        return value.lower()

    @p.field_validator("registration_date", mode="before")
    def validate_dt(cls, value: str) -> datetime.date:
        """Validate the registration date."""
        return datetime.datetime.strptime(value, "%m/%d/%Y").date()


class SearchResponseVoters(p.BaseModel, frozen=True):
    """The voters data structure for the search response."""

    values: t.Sequence[SearchResponseVoter] = p.Field(alias="$values")


class SearchResponseData(p.BaseModel, frozen=True):
    """The inner data structure for the search response."""

    voters: SearchResponseVoters


class SearchResponse(p.BaseModel, frozen=True):
    """The response structure for the search API."""

    data: SearchResponseData = p.Field(alias="Data")
    success: bool = p.Field(alias="Success")
    error_message: str | None = p.Field(alias="ErrorMessage")
    warning_message: str | None = p.Field(alias="WarningMessage")


# ------------------------------------------------------------------------
# CheckRegistrationTool implementation for WI
# ------------------------------------------------------------------------


class WisconsinCheckRegistrationTool(CheckRegistrationTool):
    """A tool for checking voter registration in Wisconsin."""

    SEARCH_URL: t.ClassVar[str] = (
        "https://myvote.wi.gov/DesktopModules/GabMyVoteModules/api/voter/search"
    )

    state: t.ClassVar[str] = "WI"
    features: t.ClassVar[SupportedFeatures] = SupportedFeatures(details=True)

    def check_registration(
        self,
        first_name: str,
        last_name: str,
        zipcode: str,
        birth_date: datetime.date,
        details: bool = False,
    ) -> CheckRegistrationResult:
        """Check if a voter is registered in Michigan."""
        try:
            request = httpx.post(
                self.SEARCH_URL,
                json={
                    "firstName": first_name,
                    "lastName": last_name,
                    "birthDate": birth_date.strftime("%m/%d/%Y"),
                },
            )
            request.raise_for_status()
        except httpx.HTTPError as e:
            raise CheckRegistrationError("Failed to check voter registration.") from e

        try:
            response = SearchResponse.model_validate(request.json())
        except Exception as e:
            raise CheckRegistrationError("Failed to parse response.") from e

        if not response.success:
            raise CheckRegistrationError(response.error_message or "Unknown error.")

        if not response.data.voters.values:
            return CheckRegistrationResult(registered=False)

        if len(list(response.data.voters.values)) > 1:
            raise MultipleRecordsFoundError()

        if not details:
            return CheckRegistrationResult(registered=True)

        voter = response.data.voters.values[0]
        return CheckRegistrationResult(
            registered=True,
            details=CheckRegistrationDetails(
                state_id=voter.voter_id,
                registration_date=voter.registration_date,
                status=voter.voter_status,
            ),
        )



#!/usr/bin/env python

import csv
import datetime
import os
import pathlib
import sys
import typing as t
from xml.etree import ElementTree as ET

import click

from . import PennsylvaniaAPIClient, get_check_tool
from .pa.debug import CurlDebugTransport


@click.group()
@click.version_option()
def vote():
    """Command-line tool for working with voter data."""
    pass


@vote.command()
@click.argument("first_name", type=str, required=True)
@click.argument("last_name", type=str, required=True)
@click.argument("zipcode", type=str, required=True)
@click.argument("birthday", type=click.DateTime(), required=True)
@click.option(
    "--details", is_flag=True, default=False, help="Return detailed information."
)
def check(
    first_name: str,
    last_name: str,
    zipcode: str,
    birthday: datetime.date,
    details: bool = False,
) -> None:
    """Check if a single person is registered to vote."""
    tool = get_check_tool(zipcode=zipcode)
    if tool is None:
        print(f"Error: unsupported state for zipcode {zipcode}.")
        sys.exit(1)
    result = tool.check_registration(first_name, last_name, zipcode, birthday, details)
    if result.registered:
        print(f"{first_name} {last_name} is registered to vote in {tool.state}.")
        if details and result.details:
            print(result.details.model_dump_json(indent=2))
    else:
        print(f"{first_name} {last_name} is not registered to vote in {tool.state}.")


@vote.command()
@click.argument("csv_path", type=click.Path(exists=True))
@click.option(
    "--details", is_flag=True, default=False, help="Return detailed information."
)
@click.option(
    "-f",
    "--first-name",
    "first_name_header",
    type=str,
    default="First Name",
    help="Name of the 'First Name' column.",
)
@click.option(
    "-l",
    "--last-name",
    "last_name_header",
    type=str,
    default="Last Name",
    help="Name of the 'Last Name' column.",
)
@click.option(
    "-d",
    "--dob",
    "dob_header",
    type=str,
    default="Date of Birth",
    help="Name of the 'Date of Birth' column.",
)
@click.option(
    "-z",
    "--zipcode",
    "zipcode_header",
    type=str,
    default="Zipcode",
    help="Name of the 'Zipcode' column.",
)
@click.option(
    "-r",
    "--registered",
    "registered_header",
    type=str,
    default="Registered",
    help="Name of the 'Registered' column.",
)
@click.option(
    "--registration-date",
    "registration_date_header",
    type=str,
    default="Registration Date",
    help="Name of the 'Registration Date' column.",
)
@click.option(
    "--registration-status",
    "registration_status_header",
    type=str,
    default="Registration Status",
    help="Name of the 'Registration Status' column.",
)
@click.option(
    "--state-voter-id",
    "state_voter_id_header",
    type=str,
    default="State Voter ID",
    help="Name of the 'State Voter ID' column.",
)
def check_csv(
    csv_path: pathlib.Path,
    details: bool = False,
    first_name_header: str = "First Name",
    last_name_header: str = "Last Name",
    dob_header: str = "Date of Birth",
    zipcode_header: str = "Zipcode",
    registered_header: str = "Registered",
    registration_date_header: str = "Registration Date",
    registration_status_header: str = "Registration Status",
    state_voter_id_header: str = "State Voter ID",
) -> None:
    """
    Check if multiple people are registered to vote.

    Output a new CSV. If `details` is False, a single column is added to the CSV
    indicating whether the person is registered to vote.

    If `details` is True, additional columns are added including registration
    date and state voter ID, if known.
    """
    extra_fields = [registered_header]
    if details:
        extra_fields += [
            registration_date_header,
            registration_status_header,
            state_voter_id_header,
        ]

    with open(csv_path, "r") as f:
        # Read the CSV file and begin the new output CSV
        reader = csv.DictReader(f)
        if not reader.fieldnames:
            raise ValueError("Invalid CSV format: missing header row.")
        field_names = list(reader.fieldnames) + extra_fields
        writer = csv.DictWriter(sys.stdout, fieldnames=field_names)
        writer.writeheader()

        for row in reader:
            # Read the relevant cells from the row
            try:
                first_name, last_name, dob_str, zipcode = (
                    row[first_name_header],
                    row[last_name_header],
                    row[dob_header],
                    row[zipcode_header],
                )
                dob = datetime.datetime.strptime(dob_str, "%Y-%m-%d").date()
            except (KeyError, ValueError) as e:
                raise ValueError(
                    "Invalid CSV format: unable to parse required fields."
                ) from e

            # Add default values for the new details columns, if needed
            if details:
                row[registration_date_header] = ""
                row[registration_status_header] = ""
                row[state_voter_id_header] = ""

            # Get the registration tool for the given ZIP code
            tool = get_check_tool(zipcode=zipcode)

            # Handle unsupported states
            if tool is None:
                row["Registered"] = "(unsupported state)"
                continue

            # Check the registration status for this voter
            result = tool.check_registration(
                first_name, last_name, zipcode, dob, details
            )

            # Finish quickly if the voter is not registered
            if not result.registered:
                row[registered_header] = False
                writer.writerow(row)
                continue

            # Update the row with the registration status and details if requested
            row[registered_header] = True
            if details and result.details:
                row[registration_date_header] = (
                    result.details.registration_date.strftime("%Y-%m-%d")
                )
                row[registration_status_header] = result.details.status
                row[state_voter_id_header] = result.details.state_id

            writer.writerow(row)


@vote.group()
def pa():
    """Commands for working with Pennsylvania voter registration."""
    pass


def _pa_api_key() -> str:
    API_KEY = os.environ.get("PA_API_KEY")
    if not API_KEY:
        raise ValueError("PA_API_KEY environment variable must be set.")
    return API_KEY


@pa.command()
def get_application_setup():
    """Get the application setup."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    setup = client.get_application_setup()
    print(setup)


@pa.command()
def get_ballot_application_setup():
    """Get the mail-in ballot setup."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    setup = client.get_ballot_application_setup()
    print(setup)


@pa.command()
def get_languages():
    """Get the languages."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    languages = client.get_languages()
    print(languages)


@pa.command()
def get_xml_template():
    """Get the XML template."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    template = client.get_xml_template()
    ET.indent(template)
    ET.dump(template)


@pa.command()
def get_ballot_xml_template():
    """Get the ballot XML template."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    template = client.get_ballot_xml_template()
    ET.indent(template)
    ET.dump(template)


@pa.command()
def get_error_values():
    """Get the error values."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    errors = client.get_error_values()
    print(errors)


@pa.command()
@click.argument("county")
def get_municipalities(county: str):
    """Get the municipalities for the given county."""
    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    municipalities = client.get_municipalities(county)
    print(municipalities)


@pa.command()
@click.option("--xml", type=click.File("r"))
@click.option("--timeout", type=float, default=100.0, help="Timeout in seconds.")
@click.option(
    "--debug-curl",
    is_flag=True,
    default=False,
    help="Display the request as a CURL command.",
)
def set_xml_application(xml: t.TextIO, timeout: float, debug_curl: bool):
    """Validate an arbitrary XML application structure and submit it if valid."""
    from .pa.client import VoterApplication

    xml_str = xml.read()
    application = VoterApplication.from_xml(xml_str)
    transport = (
        CurlDebugTransport(click.get_text_stream("stdout")) if debug_curl else None
    )
    client = PennsylvaniaAPIClient.staging(
        _pa_api_key(), timeout=timeout, _transport=transport
    )
    response = client.set_application(application)
    print(response)


@pa.command()
def set_test_application():
    """Submit a test application with test data."""
    import datetime

    from .pa.client import (
        BatchMode,
        GenderChoice,
        PoliticalPartyChoice,
        RegistrationKind,
        VoterApplication,
        VoterApplicationRecord,
    )

    ar = VoterApplicationRecord(
        batch=BatchMode.ALL_ERRORS,
        first_name="Test",
        last_name="Applicant1",
        is_us_citizen=True,
        will_be_18=True,
        political_party=PoliticalPartyChoice.DEMOCRATIC,
        gender=GenderChoice.MALE,
        email="test.applicant1@example.com",
        birth_date=datetime.date(1980, 1, 1),
        registration_kind=RegistrationKind.NEW,
        confirm_declaration=True,
        address="123 Main St",
        city="Philadelphia",
        zip5="19127",
        drivers_license="12345678",
    )
    application = VoterApplication(record=ar)

    client = PennsylvaniaAPIClient.staging(_pa_api_key())
    response = client.set_application(application)
    print(response)


if __name__ == "__main__":
    vote()



class VoterToolsError(Exception):
    """Base class for all voter tools errors."""

    pass


class CheckRegistrationError(Exception):
    """An error occurred while checking voter registration."""

    pass


class MultipleRecordsFoundError(CheckRegistrationError):
    """Multiple voters were found matching the given information."""

    pass


class APIError(Exception):
    """Base class for all API errors."""

    pass



from enum import Enum

from ..zipcodes import get_county


class CountyChoice(str, Enum):
    """
    Enum for the counties in Pennsylvania and their county codes.

    CONSIDER: This list is also returned by the API *SETUP calls, but it is
    used elsewhere (on PA's voter-facing websites, too) and it seems stable
    enough to hard-code here.
    """

    ADAMS = "ADAMS"  # 2290
    ALLEGHENY = "ALLEGHENY"  # 2291
    ARMSTRONG = "ARMSTRONG"  # 2292
    BEDFORD = "BEDFORD"  # 2294
    BERKS = "BERKS"  # 2295
    BLAIR = "BLAIR"  # 2296
    BRADFORD = "BRADFORD"  # 2297
    BUCKS = "BUCKS"  # 2298
    BUTLER = "BUTLER"  # 2299
    CAMBRIA = "CAMBRIA"  # 2300
    CAMERON = "CAMERON"  # 2301
    CARBON = "CARBON"  # 2302
    CENTRE = "CENTRE"  # 2303
    CHESTER = "CHESTER"  # 2304
    CLARION = "CLARION"  # 2305
    CLEARFIELD = "CLEARFIELD"  # 2306
    CLINTON = "CLINTON"  # 2307
    COLUMBIA = "COLUMBIA"  # 2308
    CRAWFORD = "CRAWFORD"  # 2309
    CUMBERLAND = "CUMBERLAND"  # 2310
    DAUPHIN = "DAUPHIN"  # 2311
    DELAWARE = "DELAWARE"  # 2312
    ELK = "ELK"  # 2313
    ERIE = "ERIE"  # 2314
    FAYETTE = "FAYETTE"  # 2315
    FOREST = "FOREST"  # 2316
    FRANKLIN = "FRANKLIN"  # 2317
    FULTON = "FULTON"  # 2318
    GREENE = "GREENE"  # 2319
    HUNTINGDON = "HUNTINGDON"  # 2320
    INDIANA = "INDIANA"  # 2321
    JEFFERSON = "JEFFERSON"  # 2322
    JUNIATA = "JUNIATA"  # 2323
    LACKAWANNA = "LACKAWANNA"  # 2324
    LANCASTER = "LANCASTER"  # 2325
    LAWRENCE = "LAWRENCE"  # 2326
    LEBANON = "LEBANON"  # 2327
    LEHIGH = "LEHIGH"  # 2328
    LUZERNE = "LUZERNE"  # 2329
    LYCOMING = "LYCOMING"  # 2330
    McKEAN = "McKEAN"  # 2331
    MERCER = "MERCER"  # 2332
    MIFFLIN = "MIFFLIN"  # 2333
    MONROE = "MONROE"  # 2334
    MONTGOMERY = "MONTGOMERY"  # 2335
    MONTOUR = "MONTOUR"  # 2336
    NORTHAMPTON = "NORTHAMPTON"  # 2337
    NORTHUMBERLAND = "NORTHUMBERLAND"  # 2338
    PERRY = "PERRY"  # 2339
    PHILADELPHIA = "PHILADELPHIA"  # 2340
    PIKE = "PIKE"  # 2341
    POTTER = "POTTER"  # 2342
    SCHUYLKILL = "SCHUYLKILL"  # 2343
    SNYDER = "SNYDER"  # 2344
    SOMERSET = "SOMERSET"  # 2345
    SULLIVAN = "SULLIVAN"  # 2346
    SUSQUEHANNA = "SUSQUEHANNA"  # 2347
    TIOGA = "TIOGA"  # 2348
    UNION = "UNION"  # 2349
    VENANGO = "VENANGO"  # 2350
    WARREN = "WARREN"  # 2351
    WASHINGTON = "WASHINGTON"  # 2352
    WAYNE = "WAYNE"  # 2353
    WESTMORELAND = "WESTMORELAND"  # 2354
    WYOMING = "WYOMING"  # 2355
    YORK = "YORK"  # 2356


def get_county_choice(zipcode: str) -> CountyChoice | None:
    """Get the county code for a Pennsylvania ZIP code."""
    county = get_county(zipcode)
    if county is None:
        return None
    try:
        return CountyChoice[county.upper()]
    except KeyError:
        return None



import pathlib
import typing as t
from datetime import date

import httpx
from user_agent import generate_user_agent

from ..errors import CheckRegistrationError
from ..tool import CheckRegistrationResult, CheckRegistrationTool, SupportedFeatures
from ..zipcodes import get_county

COUNTY_TO_CODE = {
    "ADAMS": "2290",
    "ALLEGHENY": "2291",
    "ARMSTRONG": "2292",
    "BEAVER": "2293",
    "BEDFORD": "2294",
    "BERKS": "2295",
    "BLAIR": "2296",
    "BRADFORD": "2297",
    "BUCKS": "2298",
    "BUTLER": "2299",
    "CAMBRIA": "2300",
    "CAMERON": "2301",
    "CARBON": "2302",
    "CENTRE": "2303",
    "CHESTER": "2304",
    "CLARION": "2305",
    "CLEARFIELD": "2306",
    "CLINTON": "2307",
    "COLUMBIA": "2308",
    "CRAWFORD": "2309",
    "CUMBERLAND": "2310",
    "DAUPHIN": "2311",
    "DELAWARE": "2312",
    "ELK": "2313",
    "ERIE": "2314",
    "FAYETTE": "2315",
    "FOREST": "2316",
    "FRANKLIN": "2317",
    "FULTON": "2318",
    "GREENE": "2319",
    "HUNTINGDON": "2320",
    "INDIANA": "2321",
    "JEFFERSON": "2322",
    "JUNIATA": "2323",
    "LACKAWANNA": "2324",
    "LANCASTER": "2325",
    "LAWRENCE": "2326",
    "LEBANON": "2327",
    "LEHIGH": "2328",
    "LUZERNE": "2329",
    "LYCOMING": "2330",
    "McKEAN": "2331",
    "MERCER": "2332",
    "MIFFLIN": "2333",
    "MONROE": "2334",
    "MONTGOMERY": "2335",
    "MONTOUR": "2336",
    "NORTHAMPTON": "2337",
    "NORTHUMBERLAND": "2338",
    "PERRY": "2339",
    "PHILADELPHIA": "2340",
    "PIKE": "2341",
    "POTTER": "2342",
    "SCHUYLKILL": "2343",
    "SNYDER": "2344",
    "SOMERSET": "2345",
    "SULLIVAN": "2346",
    "SUSQUEHANNA": "2347",
    "TIOGA": "2348",
    "UNION": "2349",
    "VENANGO": "2350",
    "WARREN": "2351",
    "WASHINGTON": "2352",
    "WAYNE": "2353",
    "WESTMORELAND": "2354",
    "WYOMING": "2355",
    "YORK": "2356",
}


def get_county_code(zipcode: str) -> str | None:
    """Get the county code for a Pennsylvania ZIP code."""
    county = get_county(zipcode)
    if not county:
        return None
    return COUNTY_TO_CODE.get(county.upper())


class PennsylvaniaCheckRegistrationTool(CheckRegistrationTool):
    """A tool for checking voter registration in Pennsylvania."""

    state: t.ClassVar[str] = "PA"
    features: t.ClassVar[SupportedFeatures] = SupportedFeatures(details=False)

    STATUS_URL: t.ClassVar[str] = (
        "https://www.pavoterservices.pa.gov/Pages/voterregistrationstatus.aspx"
    )
    ASPDATA_PATH: t.ClassVar[pathlib.Path] = (
        pathlib.Path(__file__).parent / "pa.aspdata.txt"
    )

    def _get_view_state_data(self) -> dict[str, str]:
        """Return obscure ASP.NET strucutred data. Annoying."""
        with open(self.ASPDATA_PATH) as f:
            lines = f.readlines()
        lines = [stripped for line in lines if (stripped := line.strip())]
        return {kv[0]: kv[1].strip() for line in lines if (kv := line.split(":", 1))}

    def _request(
        self, first_name: str, last_name: str, zipcode: str, birth_date: date
    ) -> httpx.Response:
        """Make a request to the PA voter registration status page."""
        county_code = get_county_code(zipcode)
        if not county_code:
            raise CheckRegistrationError("Invalid ZIP code or unknown county")
        data = {
            "ctl00$ContentPlaceHolder1$ScriptManager1": "ctl00$ContentPlaceHolder1$UpdatePanel1|ctl00$ContentPlaceHolder1$btnContinue",  # noqa: E501
            "ctl00_ContentPlaceHolder1_ScriptManager1_HiddenField": "",
            "ctl00$ContentPlaceHolder1$SearchSelection": "rdoSearchByName",
            "ctl00$ContentPlaceHolder1$CountyCombo": county_code,
            "ctl00$ContentPlaceHolder1$txtVRSzip": zipcode,
            "ctl00$ContentPlaceHolder1$txtVRSOpt2Item2": first_name,
            "ctl00$ContentPlaceHolder1$txtVRSOpt2Item3": last_name,
            "ctl00$ContentPlaceHolder1$txtVRSOpt2Item4": birth_date.strftime(
                "%m/%d/%Y"
            ),
            **self._get_view_state_data(),
            "ctl00$ContentPlaceHolder1$btnContinue": "Search",
        }
        response = httpx.post(
            self.STATUS_URL,
            data=data,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": "https://www.pavoterservices.pa.gov/pages/voterregistrationstatus.aspx",
                "Origin": "https://www.pavoterservices.pa.gov",
                "User-Agent": generate_user_agent(os="mac"),
                "X-MicrosoftAjax": "Delta=true",
                "X-Requested-With": "XMLHttpRequest",
            },
        )
        response.raise_for_status()
        return response

    def check_registration(
        self,
        first_name: str,
        last_name: str,
        zipcode: str,
        birth_date: date,
        details: bool = False,
    ) -> CheckRegistrationResult:
        """Check whether a voter is registered in Pennsylvania."""
        try:
            response = self._request(first_name, last_name, zipcode, birth_date)
        except httpx.HTTPStatusError as e:
            raise CheckRegistrationError("Failed to check voter registration") from e
        except CheckRegistrationError:
            raise

        return CheckRegistrationResult(
            registered="voter status record" in response.text.lower(), details=None
        )



"""Module for looking up zip codes, states, and counties."""

import csv
import pathlib

_ZIP_TO_STATE: dict[str, str] = {}
_ZIP_TO_COUNTY: dict[str, str] = {}

_ZIP_PATH = pathlib.Path(__file__).parent / "zipcodes.us.csv"

_ZIPCODE_COLUMN = 1
_STATE_COLUMN = 4
_COUNTY_COLUMN = 5


def _load_zipcodes():
    global _ZIP_TO_STATE, _ZIP_TO_COUNTY
    if not _ZIP_TO_STATE or not _ZIP_TO_COUNTY:
        with open(_ZIP_PATH, "r") as f:
            reader = csv.reader(f)
            next(reader)
            for row in reader:
                _ZIP_TO_STATE[row[_ZIPCODE_COLUMN]] = row[_STATE_COLUMN]
                _ZIP_TO_COUNTY[row[_ZIPCODE_COLUMN]] = row[_COUNTY_COLUMN]


def _get_zip_to_state() -> dict[str, str]:
    """Return a dictionary mapping zip codes to state abbreviations."""
    _load_zipcodes()
    return _ZIP_TO_STATE


def _get_zip_to_county() -> dict[str, str]:
    """Return a dictionary mapping zip codes to county names."""
    _load_zipcodes()
    return _ZIP_TO_COUNTY


def get_state(zipcode: str) -> str | None:
    """Return the state abbreviation for a given zip code."""
    return _get_zip_to_state().get(zipcode)


def get_county(zipcode: str) -> str | None:
    """Return the county name for a given zip code."""
    return _get_zip_to_county().get(zipcode)



import typing as t

from ..errors import APIError


class InvalidAccessKeyError(APIError):
    """
    Raised when an invalid access key is provided.

    The key may be invalid, or it may not have the necessary permissions.

    For instance, a read-only key may not be used to submit registrations.
    """

    _message: t.ClassVar[str] = """

    Invalid API access key. It may be *entirely* invalid, or it may simply not
    have the permissions necessary to fulfill the request. For instance, the
    PA SOS hands out read-only keys that can't be used to submit registrations,
    and also hands out keys that cannot be used for mail-in applications.
"""

    def __init__(self, message: str = _message) -> None:
        """Initialize the error with the given message."""
        super().__init__(message)


class UnexpectedResponseError(APIError):
    """Raised when an unexpected response is received from the server."""

    pass


class InvalidRegistrationError(APIError):
    """Raised when an invalid voter registration is provided."""

    pass


class InvalidDLError(APIError):
    """Raised when an invalid driver's license is provided."""

    pass


class InvalidSignatureError(APIError):
    """Raised when an invalid signature is provided."""

    pass


_CODE_TO_ERROR: t.Mapping[str, t.Type[APIError]] = {
    "VR_WAPI_InvalidAccessKey": InvalidAccessKeyError,
    "VR_WAPI_InvalidOVRDL": InvalidDLError,
    "VR_WAPI_Invalidsignaturestring": InvalidSignatureError,
    "VR_WAPI_Invalidsignaturetype": InvalidSignatureError,
    "VR_WAPI_Invalidsignaturesize": InvalidSignatureError,
    "VR_WAPI_Invalidsignaturedimension": InvalidSignatureError,
    "VR_WAPI_Invalidsignaturecontrast": InvalidSignatureError,
    "VR_WAPI_Invalidsignatureresolution": InvalidSignatureError,
}


def get_error_class(
    code: str, default: t.Type[APIError] = APIError
) -> t.Type[APIError]:
    """Return the error class for the given error code."""
    return _CODE_TO_ERROR.get(code, default)
